"""Command utilities for managing pgstac."""

import logging
import sys
from typing import Optional

import fire
from smart_open import open

from pypgstac.db import PgstacDB
from pypgstac.load import Loader, Methods, Tables
from pypgstac.migrate import Migrate


class PgstacCLI:
    """CLI for PgSTAC."""

    def __init__(
        self,
        dsn: Optional[str] = "",
        version: bool = False,
        debug: bool = False,
        usequeue: bool = False,
    ):
        """Initialize PgSTAC CLI."""
        if version:
            sys.exit(0)

        self.dsn = dsn
        self._db = PgstacDB(dsn=dsn, debug=debug, use_queue=usequeue)
        if debug:
            logging.basicConfig(level=logging.DEBUG)
            sys.tracebacklimit = 1000

    @property
    def initversion(self) -> str:
        """Return earliest migration version."""
        return "0.1.9"

    @property
    def version(self) -> Optional[str]:
        """Get PgSTAC version installed on database."""
        return self._db.version

    @property
    def pg_version(self) -> str:
        """Get PostgreSQL server version installed on database."""
        return self._db.pg_version

    def pgready(self) -> None:
        """Wait for a pgstac database to accept connections."""
        self._db.wait()

    def search(self, query: str) -> str:
        """Search PgSTAC."""
        return self._db.search(query)

    def migrate(self, toversion: Optional[str] = None) -> str:
        """Migrate PgSTAC Database."""
        migrator = Migrate(self._db)
        return migrator.run_migration(toversion=toversion)

    def load(
        self,
        table: Tables,
        file: str,
        method: Optional[Methods] = Methods.insert,
        dehydrated: Optional[bool] = False,
        chunksize: Optional[int] = 10000,
    ) -> None:
        """Load collections or items into PgSTAC."""
        loader = Loader(db=self._db)
        if table == "collections":
            loader.load_collections(file, method)
        if table == "items":
            loader.load_items(file, method, dehydrated, chunksize)

    def runqueue(self) -> str:
        return self._db.run_queued()

    def loadextensions(self) -> None:
        conn = self._db.connect()

        with conn.cursor() as cur:
            cur.execute(
                """
                INSERT INTO stac_extensions (url)
                SELECT DISTINCT
                substring(
                    jsonb_array_elements_text(content->'stac_extensions') FROM E'^[^#]*'
                )
                FROM collections
                ON CONFLICT DO NOTHING;
            """,
            )
            conn.commit()

        urls = self._db.query(
            """
                SELECT url FROM stac_extensions WHERE content IS NULL;
            """,
        )
        if urls:
            for u in urls:
                url = u[0]
                try:
                    with open(url, "r") as f:
                        content = f.read()
                        self._db.query(
                            """
                                UPDATE pgstac.stac_extensions
                                SET content=%s
                                WHERE url=%s
                                ;
                            """,
                            [content, url],
                        )
                        conn.commit()
                except Exception:
                    pass


def cli() -> fire.Fire:
    """Wrap fire call for CLI."""
    fire.Fire(PgstacCLI)


if __name__ == "__main__":
    fire.Fire(PgstacCLI)



"""Test Hydration."""

import json
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, cast

from pypgstac import hydration
from pypgstac.hydration import DO_NOT_MERGE_MARKER
from pypgstac.load import Loader

HERE = Path(__file__).parent
LANDSAT_COLLECTION = (
    HERE / ".." / "data-files" / "hydration" / "collections" / "landsat-c2-l1.json"
)
LANDSAT_DEHYDRATED_ITEM = (
    HERE
    / ".."
    / "data-files"
    / "hydration"
    / "dehydrated-items"
    / "landsat-c2-l1"
    / "LM04_L1GS_001001_19830527_02_T2.json"
)

LANDSAT_ITEM = (
    HERE
    / ".."
    / "data-files"
    / "hydration"
    / "raw-items"
    / "landsat-c2-l1"
    / "LM04_L1GS_001001_19830527_02_T2.json"
)


class TestHydrate:
    def hydrate(
        self,
        base_item: Dict[str, Any],
        item: Dict[str, Any],
    ) -> Dict[str, Any]:
        hpy = hydration.hydrate_py(deepcopy(base_item), deepcopy(item))
        hrs = hydration.hydrate(deepcopy(base_item), deepcopy(item))
        assert hpy == hrs
        return hrs

    def test_landsat_c2_l1(self, loader: Loader) -> None:
        """Test that a dehydrated item is is equal to the raw item it was dehydrated
        from, against the base item of the collection
        .
        """
        with open(LANDSAT_COLLECTION) as f:
            collection = json.load(f)
        loader.load_collections(str(LANDSAT_COLLECTION))

        with open(LANDSAT_DEHYDRATED_ITEM) as f:
            dehydrated = json.load(f)

        with open(LANDSAT_ITEM) as f:
            raw_item = json.load(f)

        base_item = cast(
            Dict[str, Any],
            loader.db.query_one(
                "SELECT base_item FROM collections WHERE id=%s;",
                (collection["id"],),
            ),
        )

        assert type(base_item) == dict

        hydrated = self.hydrate(base_item, dehydrated)
        assert hydrated == raw_item

    def test_full_hydrate(self) -> None:
        base_item = {"a": "first", "b": "second", "c": "third"}
        dehydrated: Dict[str, Any] = {}

        rehydrated = self.hydrate(base_item, dehydrated)
        assert rehydrated == base_item

    def test_full_nested(self) -> None:
        base_item = {"a": "first", "b": "second", "c": {"d": "third"}}
        dehydrated: Dict[str, Any] = {}

        rehydrated = self.hydrate(base_item, dehydrated)
        assert rehydrated == base_item

    def test_nested_extra_keys(self) -> None:
        """
        Test that items having nested dicts with keys not in base item preserve
        the additional keys in the dehydrated item.
        """
        base_item = {"a": "first", "b": "second", "c": {"d": "third"}}
        dehydrated = {"c": {"e": "fourth", "f": "fifth"}}
        hydrated = self.hydrate(base_item, dehydrated)

        assert hydrated == {
            "a": "first",
            "b": "second",
            "c": {"d": "third", "e": "fourth", "f": "fifth"},
        }

    def test_list_of_dicts_extra_keys(self) -> None:
        """Test that an equal length list of dicts is hydrated correctly."""
        base_item = {"a": [{"b1": 1, "b2": 2}, {"c1": 1, "c2": 2}]}
        dehydrated = {"a": [{"b3": 3}, {"c3": 3}]}

        hydrated = self.hydrate(base_item, dehydrated)
        assert hydrated == {
            "a": [{"b1": 1, "b2": 2, "b3": 3}, {"c1": 1, "c2": 2, "c3": 3}],
        }

    def test_equal_len_list_of_mixed_types(self) -> None:
        """
        Test that a list of equal length containing matched types at
        each index dehydrates
        dicts and preserves item-values of other types.
        """
        base_item = {"a": [{"b1": 1, "b2": 2}, "foo", {"c1": 1, "c2": 2}, "bar"]}
        dehydrated = {"a": [{"b3": 3}, "far", {"c3": 3}, "boo"]}

        hydrated = self.hydrate(base_item, dehydrated)
        assert hydrated == {
            "a": [
                {"b1": 1, "b2": 2, "b3": 3},
                "far",
                {"c1": 1, "c2": 2, "c3": 3},
                "boo",
            ],
        }

    def test_unequal_len_list(self) -> None:
        """Test that unequal length lists preserve the item value exactly."""
        base_item = {"a": [{"b1": 1}, {"c1": 1}, {"d1": 1}]}
        dehydrated = {"a": [{"b1": 1, "b2": 2}, {"c1": 1, "c2": 2}]}

        hydrated = self.hydrate(base_item, dehydrated)
        assert hydrated == dehydrated

    def test_marked_non_merged_fields(self) -> None:
        base_item = {
            "a": "first",
            "b": "second",
            "c": {"d": "third", "e": "fourth"},
        }
        dehydrated = {"c": {"e": DO_NOT_MERGE_MARKER, "f": "fifth"}}

        hydrated = self.hydrate(base_item, dehydrated)
        assert hydrated == {
            "a": "first",
            "b": "second",
            "c": {"d": "third", "f": "fifth"},
        }

    def test_marked_non_merged_fields_in_list(self) -> None:
        base_item = {
            "a": [{"b": "first", "d": "third"}, {"c": "second", "e": "fourth"}],
        }
        dehydrated = {
            "a": [
                {"d": DO_NOT_MERGE_MARKER},
                {"e": DO_NOT_MERGE_MARKER, "f": "fifth"},
            ],
        }

        hydrated = self.hydrate(base_item, dehydrated)
        assert hydrated == {"a": [{"b": "first"}, {"c": "second", "f": "fifth"}]}

    def test_deeply_nested_dict(self) -> None:
        base_item = {"a": {"b": {"c": {"d": "first", "d1": "second"}}}}
        dehydrated = {"a": {"b": {"c": {"d2": "third"}}}}

        hydrated = self.hydrate(base_item, dehydrated)
        assert hydrated == {
            "a": {"b": {"c": {"d": "first", "d1": "second", "d2": "third"}}},
        }

    def test_equal_list_of_non_dicts(self) -> None:
        """Values of lists that match base_item should be hydrated back on."""
        base_item = {"assets": {"thumbnail": {"roles": ["thumbnail"]}}}
        dehydrated = {"assets": {"thumbnail": {"href": "http://foo.com"}}}

        hydrated = self.hydrate(base_item, dehydrated)
        assert hydrated == {
            "assets": {"thumbnail": {"roles": ["thumbnail"], "href": "http://foo.com"}},
        }

    def test_invalid_assets_removed(self) -> None:
        """
        Assets can be included on item-assets that are not uniformly included on
        individual items. Ensure that item asset keys from base_item aren't included
        after hydration.
        """
        base_item = {
            "type": "Feature",
            "assets": {
                "asset1": {"name": "Asset one"},
                "asset2": {"name": "Asset two"},
            },
        }

        dehydrated = {
            "assets": {
                "asset1": {"href": "http://foo.com"},
                "asset2": DO_NOT_MERGE_MARKER,
            },
        }

        hydrated = self.hydrate(base_item, dehydrated)

        assert hydrated == {
            "type": "Feature",
            "assets": {"asset1": {"name": "Asset one", "href": "http://foo.com"}},
        }

    def test_top_level_base_keys_marked(self) -> None:
        """
        Top level keys on the base item not present on the incoming item should
        be marked as do not merge, no matter the nesting level.
        """
        base_item = {
            "single": "Feature",
            "double": {"nested": "value"},
            "triple": {"nested": {"deep": "value"}},
            "included": "value",
        }

        dehydrated = {
            "single": DO_NOT_MERGE_MARKER,
            "double": DO_NOT_MERGE_MARKER,
            "triple": DO_NOT_MERGE_MARKER,
            "unique": "value",
        }

        hydrated = self.hydrate(base_item, dehydrated)

        assert hydrated == {"included": "value", "unique": "value"}

    def test_base_none(self) -> None:
        base_item = {"value": None}
        dehydrated = {"value": {"a": "b"}}
        hydrated = self.hydrate(base_item, dehydrated)
        assert hydrated == {"value": {"a": "b"}}



"""Base library for database interaction with PgSTAC."""
import atexit
import logging
import time
from types import TracebackType
from typing import Any, Generator, List, Optional, Tuple, Type, Union

import orjson
import psycopg
from psycopg import Connection, sql
from psycopg.types.json import set_json_dumps, set_json_loads
from psycopg_pool import ConnectionPool

try:
    from pydantic.v1 import BaseSettings  # type:ignore
except ImportError:
    from pydantic import BaseSettings  # type:ignore

from tenacity import retry, retry_if_exception_type, stop_after_attempt

logger = logging.getLogger(__name__)


def dumps(data: dict) -> str:
    """Dump dictionary as string."""
    return orjson.dumps(data).decode()


set_json_dumps(dumps)
set_json_loads(orjson.loads)


def pg_notice_handler(notice: psycopg.errors.Diagnostic) -> None:
    """Add PG messages to logging."""
    msg = f"{notice.severity} - {notice.message_primary}"
    logger.info(msg)


class Settings(BaseSettings):
    """Base Settings for Database Connection."""

    db_min_conn_size: int = 0
    db_max_conn_size: int = 1
    db_max_queries: int = 5
    db_max_idle: int = 5
    db_num_workers: int = 1
    db_retries: int = 3

    model_config = {"env_file": ".env", "extra": "ignore"}


settings = Settings()


class PgstacDB:
    """Base class for interacting with PgSTAC Database."""

    def __init__(
        self,
        dsn: Optional[str] = "",
        pool: Optional[ConnectionPool] = None,
        connection: Optional[Connection] = None,
        commit_on_exit: bool = True,
        debug: bool = False,
        use_queue: bool = False,
    ) -> None:
        """Initialize Database."""
        self.dsn: str
        if dsn is not None:
            self.dsn = dsn
        else:
            self.dsn = ""
        self.pool = pool
        self.connection = connection
        self.commit_on_exit = commit_on_exit
        self.initial_version = "0.1.9"
        self.debug = debug
        self.use_queue = use_queue
        if self.debug:
            logging.basicConfig(level=logging.DEBUG)

    def get_pool(self) -> ConnectionPool:
        """Get Database Pool."""
        if self.pool is None:
            self.pool = ConnectionPool(
                conninfo=self.dsn,
                min_size=settings.db_min_conn_size,
                max_size=settings.db_max_conn_size,
                max_waiting=settings.db_max_queries,
                max_idle=settings.db_max_idle,
                num_workers=settings.db_num_workers,
            )
        return self.pool

    def open(self) -> None:
        """Open database pool connection."""
        self.get_pool()

    def close(self) -> None:
        """Close database pool connection."""
        if self.pool is not None:
            self.pool.close()

    def connect(self) -> Connection:
        """Return database connection."""
        pool = self.get_pool()
        if self.connection is None:
            self.connection = pool.getconn()
            self.connection.autocommit = True
            if self.debug:
                self.connection.add_notice_handler(pg_notice_handler)
                self.connection.execute(
                    "SET CLIENT_MIN_MESSAGES TO NOTICE;",
                    prepare=False,
                )
            if self.use_queue:
                self.connection.execute(
                    "SET pgstac.use_queue TO TRUE;",
                    prepare=False,
                )
            atexit.register(self.disconnect)
            self.connection.execute(
                """
                    SELECT
                        CASE
                        WHEN
                        current_setting('search_path', false) ~* '\\mpgstac\\M'
                        THEN current_setting('search_path', false)
                        ELSE set_config(
                            'search_path',
                            'pgstac,' || current_setting('search_path', false),
                            false
                            )
                        END
                    ;
                    SET application_name TO 'pgstac';
                """,
                prepare=False,
            )
        return self.connection

    def wait(self) -> None:
        """Block until database connection is ready."""
        cnt: int = 0
        while cnt < 60:
            try:
                self.connect()
                self.query("SELECT 1;")
                return None
            except psycopg.errors.OperationalError:
                time.sleep(1)
                cnt += 1
        raise psycopg.errors.CannotConnectNow

    def disconnect(self) -> None:
        """Disconnect from database."""
        try:
            if self.connection is not None:
                if self.commit_on_exit:
                    self.connection.commit()
                else:
                    self.connection.rollback()
        except Exception:
            pass
        try:
            if self.pool is not None and self.connection is not None:
                self.pool.putconn(self.connection)
        except Exception:
            pass

        self.connection = None
        self.pool = None

    def __enter__(self) -> Any:
        """Enter used for context."""
        self.connect()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Exit used for context."""
        self.disconnect()

    @retry(
        stop=stop_after_attempt(settings.db_retries),
        retry=retry_if_exception_type(psycopg.errors.OperationalError),
        reraise=True,
    )
    def query(
        self,
        query: Union[str, sql.Composed],
        args: Optional[List[Any]] = None,
        row_factory: psycopg.rows.BaseRowFactory = psycopg.rows.tuple_row,
    ) -> Generator:
        """Query the database with parameters."""
        conn = self.connect()
        try:
            with conn.cursor(row_factory=row_factory) as cursor:
                if args is None:
                    rows = cursor.execute(query, prepare=False)
                else:
                    rows = cursor.execute(query, args)
                if rows:
                    for row in rows:
                        yield row
                else:
                    yield None
        except psycopg.errors.OperationalError as e:
            # If we get an operational error check the pool and retry
            logger.warning(f"OPERATIONAL ERROR: {e}")
            if self.pool is None:
                self.get_pool()
            else:
                self.pool.check()
            raise e
        except psycopg.errors.DatabaseError as e:
            if conn is not None:
                conn.rollback()
            raise e

    def query_one(self, *args: Any, **kwargs: Any) -> Union[Tuple, str, None]:
        """Return results from a query that returns a single row."""
        try:
            r = next(self.query(*args, **kwargs))
        except StopIteration:
            return None

        if r is None:
            return None
        if len(r) == 1:
            return r[0]
        return r

    def run_queued(self) -> str:
        try:
            self.connect().execute("""
                CALL run_queued_queries();
            """)
            return "Ran Queued Queries"
        except Exception as e:
            return f"Error Running Queued Queries: {e}"

    @property
    def version(self) -> Optional[str]:
        """Get the current version number from a pgstac database."""
        try:
            version = self.query_one(
                """
                SELECT version from pgstac.migrations
                order by datetime desc, version desc limit 1;
                """,
            )
            logger.debug(f"VERSION: {version}")
            if isinstance(version, bytes):
                version = version.decode()
            if isinstance(version, str):
                return version
        except psycopg.errors.UndefinedTable:
            logger.debug("PgSTAC is not installed.")
            if self.connection is not None:
                self.connection.rollback()
        return None

    @property
    def pg_version(self) -> str:
        """Get the current pg version number from a pgstac database."""
        version = self.query_one(
            """
            SHOW server_version;
            """,
        )
        logger.debug(f"PG VERSION: {version}.")
        if isinstance(version, bytes):
            version = version.decode()
        if isinstance(version, str):
            if int(version.split(".")[0]) < 13:
                raise Exception("PgSTAC requires PostgreSQL 13+")
            return version
        else:
            if self.connection is not None:
                self.connection.rollback()
            raise Exception("Could not find PG version.")

    def func(self, function_name: str, *args: Any) -> Generator:
        """Call a database function."""
        placeholders = sql.SQL(", ").join(sql.Placeholder() * len(args))
        func = sql.Identifier(function_name)
        cleaned_args = []
        for arg in args:
            if isinstance(arg, dict):
                cleaned_args.append(psycopg.types.json.Jsonb(arg))
            else:
                cleaned_args.append(arg)
        base_query = sql.SQL("SELECT * FROM {}({});").format(func, placeholders)
        return self.query(base_query, cleaned_args)

    def search(self, query: Union[dict, str, psycopg.types.json.Jsonb] = "{}") -> str:
        """Search PgSTAC."""
        return dumps(next(self.func("search", query))[0])



"""Utilities to help migrate pgstac schema."""
import glob
import logging
import os
import re
from collections import defaultdict
from typing import Any, Dict, Iterator, List, Optional

from smart_open import open

from . import __version__
from .db import PgstacDB

dirname = os.path.dirname(__file__)
migrations_dir = os.path.join(dirname, "migrations")

logger = logging.getLogger(__name__)


class MigrationPath:
    """Calculate path from migration files to get from one version to the next."""

    def __init__(self, path: str, f: str, t: str) -> None:
        """Initialize MigrationPath."""
        self.path = path
        if f is None:
            f = "init"
        if t is None:
            raise Exception('Must set "to" version')
        if f == t:
            raise Exception("No Migration Necessary")

        self.f = f
        self.t = t

    def parse_filename(self, filename: str) -> List[str]:
        """Get version numbers from filename."""
        filename = os.path.splitext(os.path.basename(filename))[0].replace(
            "pgstac.", "",
        )
        return filename.split("-")

    def get_files(self) -> Iterator[str]:
        """Find all migration files available."""
        path = self.path.rstrip("/")
        return glob.iglob(f"{path}/*.sql")

    def build_graph(self) -> Dict:
        """Build a graph to get from one version to another."""
        graph = defaultdict(list)
        for file in self.get_files():
            parts = self.parse_filename(file)
            if len(parts) == 2:
                graph[parts[0]].append(parts[1])
            else:
                graph["init"].append(parts[0])
        return graph

    def build_path(self) -> Optional[List[str]]:
        """Create the path of ordered files needed to migrate."""
        graph = self.build_graph()
        explored: List = []
        q = [[self.f]]

        while q:
            path = q.pop(0)
            node = path[-1]
            if node not in explored:
                neighbours = graph[node]
                for neighbour in neighbours:
                    new_path = list(path)
                    new_path.append(neighbour)
                    q.append(new_path)
                    if neighbour == self.t:
                        return new_path
                explored.append(node)
        return None

    def migrations(self) -> List[str]:
        """Return the list of migrations needed in order."""
        path = self.build_path()
        if path is None:
            raise Exception(
                f"Could not determine path to get from {self.f} to {self.t}.",
            )
        if len(path) == 1:
            return [f"pgstac.{path[0]}.sql"]
        files = []
        for idx in range(len(path) - 1):
            f = f"pgstac.{path[idx]}-{path[idx+1]}.sql"
            f = f.replace("--init", "")
            files.append(f"pgstac.{path[idx]}-{path[idx+1]}.sql")
        return files


def get_sql(file: str) -> str:
    """Get sql from a file as a string."""
    sqlstrs = []
    file = re.sub("[0-9]+[.][0-9]+[.][0-9]+-dev","unreleased",file)
    fp = os.path.join(migrations_dir, file)
    file_handle: Any = open(fp)

    with file_handle as fd:
        sqlstrs.extend(fd.readlines())
    return "\n".join(sqlstrs)


class Migrate:
    """Utilities for migrating pgstac database."""

    def __init__(self, db: PgstacDB, schema: str = "pgstac"):
        """Prepare for migration."""
        self.db = db
        self.schema = schema

    def run_migration(self, toversion: Optional[str] = None) -> str:
        """Migrate a pgstac database to current version."""
        if toversion is None:
            toversion = __version__
        files = []
        if re.search(r"-dev$",toversion):
            logger.info("using unreleased version")
            toversion = "unreleased"

        pg_version = self.db.pg_version
        logger.info(f"Migrating PgSTAC on PostgreSQL Version {pg_version}")
        oldversion = self.db.version
        if oldversion == toversion:
            logger.info(f"Target database already at version: {toversion}")
            return toversion
        if oldversion is None:
            logger.info(f"No pgstac version set, installing {toversion} from scratch.")
            files.append(os.path.join(migrations_dir, f"pgstac.{toversion}.sql"))
        else:
            logger.info(f"Migrating from {oldversion} to {toversion}.")
            m = MigrationPath(migrations_dir, oldversion, toversion)
            files = m.migrations()

        if len(files) < 1:
            raise Exception("Could not find migration files")

        conn = self.db.connect()

        with conn.cursor() as cur:
            conn.autocommit = False
            for file in files:
                logger.debug(f"Running migration file {file}.")
                migration_sql = get_sql(file)
                cur.execute(migration_sql)
                logger.debug(cur.statusmessage)
                logger.debug(cur.rowcount)

            logger.debug(f"Database migrated to {toversion}")

        newversion = self.db.version
        if conn is not None:
            if newversion == toversion:
                conn.commit()
            else:
                conn.rollback()
                raise Exception(
                    "Migration failed, database rolled back to previous state.",
                )

        logger.debug(f"New Version: {newversion}")

        return newversion



import os
from contextlib import contextmanager
from typing import Any, Dict, Generator

import psycopg

from pypgstac.db import PgstacDB
from pypgstac.migrate import Migrate

from .test_dehydrate import TestDehydrate as TDehydrate


class TestDehydratePG(TDehydrate):
    """Class to test Dehydration using pgstac."""

    @contextmanager
    def db(self) -> Generator:
        """Set up database connection."""
        origdb: str = os.getenv("PGDATABASE", "")
        with psycopg.connect(autocommit=True) as conn:
            try:
                conn.execute("CREATE DATABASE pgstactestdb;")
            except psycopg.errors.DuplicateDatabase:
                pass

        os.environ["PGDATABASE"] = "pgstactestdb"

        pgdb = PgstacDB()
        with psycopg.connect(autocommit=True) as conn:
            conn.execute("DROP SCHEMA IF EXISTS pgstac CASCADE;")
        Migrate(pgdb).run_migration()

        yield pgdb

        pgdb.close()
        os.environ["PGDATABASE"] = origdb

    def dehydrate(
        self, base_item: Dict[str, Any], item: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Dehydrate item using pgstac."""
        with self.db() as db:
            return next(db.func("strip_jsonb", item, base_item))[0]



import json
from pathlib import Path
from typing import Any, Dict, cast

from pypgstac.load import Loader

HERE = Path(__file__).parent
LANDSAT_COLLECTION = (
    HERE / ".." / "data-files" / "hydration" / "collections" / "landsat-c2-l1.json"
)


def test_landsat_c2_l1(loader: Loader) -> None:
    """Test that a base item is created when a collection is loaded and that it
    is equal to the item_assets of the collection
    .
    """
    with open(LANDSAT_COLLECTION) as f:
        collection = json.load(f)
    loader.load_collections(str(LANDSAT_COLLECTION))

    base_item = cast(
        Dict[str, Any],
        loader.db.query_one(
            "SELECT base_item FROM collections WHERE id=%s;", (collection["id"],),
        ),
    )

    assert type(base_item) == dict
    assert base_item["collection"] == collection["id"]
    assert base_item["assets"] == collection["item_assets"]



"""Tests for pypgstac."""
import json
import re
from pathlib import Path
from unittest import mock

import pytest
from psycopg.errors import UniqueViolation
from version_parser import Version as V

from pypgstac.load import Loader, Methods, __version__, read_json

HERE = Path(__file__).parent
TEST_DATA_DIR = HERE.parent.parent / "pgstac" / "tests" / "testdata"
TEST_COLLECTIONS_JSON = TEST_DATA_DIR / "collections.json"
TEST_COLLECTIONS = TEST_DATA_DIR / "collections.ndjson"
TEST_ITEMS = TEST_DATA_DIR / "items.ndjson"
TEST_DEHYDRATED_ITEMS = TEST_DATA_DIR / "items.pgcopy"

S1_GRD_COLLECTION = (
    HERE / "data-files" / "hydration" / "collections" / "sentinel-1-grd.json"
)

S1_GRD_ITEM = (
    HERE
    / "data-files"
    / "hydration"
    / "raw-items"
    / "sentinel-1-grd"
    / "S1A_IW_GRDH_1SDV_20220428T034417_20220428T034442_042968_05213C.json"
)


def version_increment(source_version: str) -> str:
    source_version = re.sub("-dev$","",source_version)
    version = V(source_version)
    return ".".join(
        map(
            str,
            [
                version.get_major_version(),
                version.get_minor_version(),
                version.get_patch_version() + 1,
            ],
        ),
    )


def test_load_collections_succeeds(loader: Loader) -> None:
    """Test pypgstac collections loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS),
        insert_mode=Methods.insert,
    )


def test_load_collections_json_succeeds(loader: Loader) -> None:
    """Test pypgstac collections loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS_JSON),
        insert_mode=Methods.insert,
    )


def test_load_collections_json_duplicates_fails(loader: Loader) -> None:
    """Test pypgstac collections loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS_JSON),
        insert_mode=Methods.insert,
    )
    with pytest.raises(UniqueViolation):
        loader.load_collections(
            str(TEST_COLLECTIONS_JSON),
            insert_mode=Methods.insert,
        )


def test_load_collections_json_duplicates_with_upsert(loader: Loader) -> None:
    """Test pypgstac collections loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS_JSON),
        insert_mode=Methods.insert,
    )
    loader.load_collections(
        str(TEST_COLLECTIONS_JSON),
        insert_mode=Methods.upsert,
    )


def test_load_collections_json_duplicates_with_ignore(loader: Loader) -> None:
    """Test pypgstac collections loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS_JSON),
        insert_mode=Methods.insert,
    )
    loader.load_collections(
        str(TEST_COLLECTIONS_JSON),
        insert_mode=Methods.ignore,
    )


def test_load_items_duplicates_fails(loader: Loader) -> None:
    """Test pypgstac collections loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS),
        insert_mode=Methods.insert,
    )
    loader.load_items(
        str(TEST_ITEMS),
        insert_mode=Methods.insert,
    )

    with pytest.raises(UniqueViolation):
        loader.load_items(
            str(TEST_ITEMS),
            insert_mode=Methods.insert,
        )


def test_load_items_succeeds(loader: Loader) -> None:
    """Test pypgstac items loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS),
        insert_mode=Methods.upsert,
    )

    loader.load_items(
        str(TEST_ITEMS),
        insert_mode=Methods.insert,
    )


def test_load_items_ignore_succeeds(loader: Loader) -> None:
    """Test pypgstac items ignore loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS),
        insert_mode=Methods.ignore,
    )

    loader.load_items(
        str(TEST_ITEMS),
        insert_mode=Methods.insert,
    )

    loader.load_items(
        str(TEST_ITEMS),
        insert_mode=Methods.ignore,
    )


def test_load_items_upsert_succeeds(loader: Loader) -> None:
    """Test pypgstac items ignore loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS),
        insert_mode=Methods.ignore,
    )

    loader.load_items(
        str(TEST_ITEMS),
        insert_mode=Methods.insert,
    )

    loader.load_items(
        str(TEST_ITEMS),
        insert_mode=Methods.upsert,
    )


def test_load_items_delsert_succeeds(loader: Loader) -> None:
    """Test pypgstac items ignore loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS),
        insert_mode=Methods.ignore,
    )

    loader.load_items(
        str(TEST_ITEMS),
        insert_mode=Methods.insert,
    )

    loader.load_items(
        str(TEST_ITEMS),
        insert_mode=Methods.delsert,
    )


def test_partition_loads_default(loader: Loader) -> None:
    """Test pypgstac items ignore loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS_JSON),
        insert_mode=Methods.ignore,
    )

    loader.load_items(
        str(TEST_ITEMS),
        insert_mode=Methods.insert,
    )

    partitions = loader.db.query_one(
        """
        SELECT count(*) from partitions;
    """,
    )

    assert partitions == 1


def test_partition_loads_month(loader: Loader) -> None:
    """Test pypgstac items ignore loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS_JSON),
        insert_mode=Methods.ignore,
    )
    if loader.db.connection is not None:
        loader.db.connection.execute(
            """
            UPDATE collections SET partition_trunc='month';
        """,
        )

    loader.load_items(
        str(TEST_ITEMS),
        insert_mode=Methods.insert,
    )

    partitions = loader.db.query_one(
        """
        SELECT count(*) from partitions;
    """,
    )

    assert partitions == 2


def test_partition_loads_year(loader: Loader) -> None:
    """Test pypgstac items ignore loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS_JSON),
        insert_mode=Methods.ignore,
    )
    if loader.db.connection is not None:
        loader.db.connection.execute(
            """
            UPDATE collections SET partition_trunc='year';
        """,
        )

    loader.load_items(
        str(TEST_ITEMS),
        insert_mode=Methods.insert,
    )

    partitions = loader.db.query_one(
        """
        SELECT count(*) from partitions;
    """,
    )

    assert partitions == 1


def test_load_items_dehydrated_ignore_succeeds(loader: Loader) -> None:
    """Test pypgstac items ignore loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS),
        insert_mode=Methods.ignore,
    )

    loader.load_items(
        str(TEST_DEHYDRATED_ITEMS), insert_mode=Methods.insert, dehydrated=True,
    )

    loader.load_items(
        str(TEST_DEHYDRATED_ITEMS), insert_mode=Methods.ignore, dehydrated=True,
    )


def test_format_items_keys(loader: Loader) -> None:
    """Test pypgstac items ignore loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS_JSON),
        insert_mode=Methods.ignore,
    )

    items_iter = read_json(str(TEST_ITEMS))
    item_json = next(iter(items_iter))
    out = loader.format_item(item_json)

    # Top level keys expected after format
    assert "id" in out
    assert "collection" in out
    assert "geometry" in out
    assert "content" in out

    # Special keys expected not to be in the item content
    content_json = json.loads(out["content"])
    assert "id" not in content_json
    assert "collection" not in content_json
    assert "geometry" not in content_json

    # Ensure bbox is included in content
    assert "bbox" in content_json


def test_s1_grd_load_and_query(loader: Loader) -> None:
    """Test pypgstac items ignore loader."""
    loader.load_collections(
        str(S1_GRD_COLLECTION),
        insert_mode=Methods.ignore,
    )

    loader.load_items(str(S1_GRD_ITEM), insert_mode=Methods.insert)

    search_body = {
        "filter-lang": "cql2-json",
        "filter": {
            "op": "and",
            "args": [
                {
                    "op": "=",
                    "args": [{"property": "collection"}, "sentinel-1-grd"],
                },
                {
                    "op": "=",
                    "args": [
                        {"property": "id"},
                        "S1A_IW_GRDH_1SDV_20220428T034417_20220428T034442_042968_05213C",  # noqa: E501
                    ],
                },
            ],
        },
    }

    res = next(
        loader.db.func(
            "search",
            search_body,
        ),
    )[0]
    res["features"][0]


def test_load_dehydrated(loader: Loader) -> None:
    """Test loader for items dumped directly out of item table."""
    collections = [
        HERE / "data-files" / "hydration" / "collections" / "chloris-biomass.json",
    ]

    for collection in collections:
        loader.load_collections(
            str(collection),
            insert_mode=Methods.ignore,
        )

    dehydrated_items = HERE / "data-files" / "load" / "dehydrated.txt"

    loader.load_items(
        str(dehydrated_items), insert_mode=Methods.insert, dehydrated=True,
    )


def test_load_collections_incompatible_version(loader: Loader) -> None:
    """Test pypgstac collections loader raises an exception for incompatible version."""
    with mock.patch(
        "pypgstac.db.PgstacDB.version", new_callable=mock.PropertyMock,
    ) as mock_version:
        mock_version.return_value = "dummy"
        with pytest.raises(ValueError):
            loader.load_collections(
                str(TEST_COLLECTIONS_JSON),
                insert_mode=Methods.insert,
            )


def test_load_items_incompatible_version(loader: Loader) -> None:
    """Test pypgstac items loader raises an exception for incompatible version."""
    loader.load_collections(
        str(TEST_COLLECTIONS_JSON),
        insert_mode=Methods.insert,
    )
    with mock.patch(
        "pypgstac.db.PgstacDB.version", new_callable=mock.PropertyMock,
    ) as mock_version:
        mock_version.return_value = "dummy"
        with pytest.raises(ValueError):
            loader.load_items(
                str(TEST_ITEMS),
                insert_mode=Methods.insert,
            )


def test_load_compatible_major_minor_version(loader: Loader) -> None:
    """Test pypgstac loader doesn't raise an exception."""
    with mock.patch(
        "pypgstac.load.__version__", version_increment(__version__),
    ) as mock_version:
        loader.load_collections(
            str(TEST_COLLECTIONS_JSON),
            insert_mode=Methods.insert,
        )
        loader.load_items(
            str(TEST_ITEMS),
            insert_mode=Methods.insert,
        )
        assert mock_version != loader.db.version


def test_load_items_nopartitionconstraint_succeeds(loader: Loader) -> None:
    """Test pypgstac items loader."""
    loader.load_collections(
        str(TEST_COLLECTIONS),
        insert_mode=Methods.upsert,
    )
    loader.load_items(
        str(TEST_ITEMS),
        insert_mode=Methods.insert,
    )

    cdtmin = loader.db.query_one(
        """
        SELECT lower(constraint_dtrange)::text
        FROM partition_sys_meta WHERE partition = '_items_1';
        """,
    )

    assert cdtmin == "2011-07-31 00:00:00+00"
    with loader.db.connect() as conn:
        conn.execute(
            """
            ALTER TABLE _items_1 DROP CONSTRAINT _items_1_dt;
            """,
        )
    cdtmin = loader.db.query_one(
        """
        SELECT lower(constraint_dtrange)::text
        FROM partition_sys_meta WHERE partition = '_items_1';
        """,
    )
    assert cdtmin == "-infinity"

    loader.load_items(
        str(TEST_ITEMS),
        insert_mode=Methods.upsert,
    )
    cdtmin = loader.db.query_one(
        """
        SELECT lower(constraint_dtrange)::text
        FROM partition_sys_meta WHERE partition = '_items_1';
        """,
    )
    assert cdtmin == "2011-07-31 00:00:00+00"



import json
from pathlib import Path
from typing import Any, Dict, cast

from pypgstac import hydration
from pypgstac.hydration import DO_NOT_MERGE_MARKER
from pypgstac.load import Loader

HERE = Path(__file__).parent
LANDSAT_COLLECTION = (
    HERE / ".." / "data-files" / "hydration" / "collections" / "landsat-c2-l1.json"
)
LANDSAT_ITEM = (
    HERE
    / ".."
    / "data-files"
    / "hydration"
    / "raw-items"
    / "landsat-c2-l1"
    / "LM04_L1GS_001001_19830527_02_T2.json"
)


class TestDehydrate:
    def dehydrate(
        self, base_item: Dict[str, Any], item: Dict[str, Any],
    ) -> Dict[str, Any]:
        return hydration.dehydrate(base_item, item)

    def test_landsat_c2_l1(self, loader: Loader) -> None:
        """
        Test that a dehydrated item is created properly from a raw item against a
        base item from a collection.
        """
        with open(LANDSAT_COLLECTION) as f:
            collection = json.load(f)
        loader.load_collections(str(LANDSAT_COLLECTION))

        with open(LANDSAT_ITEM) as f:
            item = json.load(f)

        base_item = cast(
            Dict[str, Any],
            loader.db.query_one(
                "SELECT base_item FROM collections WHERE id=%s;", (collection["id"],),
            ),
        )

        assert type(base_item) == dict

        dehydrated = self.dehydrate(base_item, item)

        # Expect certain keys on base and not on dehydrated
        only_base_keys = ["type", "collection", "stac_version"]
        assert all(k in base_item for k in only_base_keys)
        assert not any(k in dehydrated for k in only_base_keys)

        # Expect certain keys on dehydrated and not on base
        only_dehydrated_keys = ["id", "bbox", "geometry", "properties"]
        assert not any(k in base_item for k in only_dehydrated_keys)
        assert all(k in dehydrated for k in only_dehydrated_keys)

        # Properties, links should be exactly the same pre- and post-dehydration
        assert item["properties"] == dehydrated["properties"]
        assert item["links"] == dehydrated["links"]

        # Check specific assets are dehydrated correctly
        thumbnail = dehydrated["assets"]["thumbnail"]
        assert list(thumbnail.keys()) == ["href"]
        assert thumbnail["href"] == item["assets"]["thumbnail"]["href"]

        # Red asset raster bands have additional `scale` and `offset` keys
        red = dehydrated["assets"]["red"]
        assert list(red.keys()) == ["href", "eo:bands", "raster:bands"]
        assert len(red["raster:bands"]) == 1
        assert list(red["raster:bands"][0].keys()) == ["scale", "offset"]
        item_red_rb = item["assets"]["red"]["raster:bands"][0]
        assert red["raster:bands"] == [
            {"scale": item_red_rb["scale"], "offset": item_red_rb["offset"]},
        ]

        # nir09 asset raster bands does not have a `unit` attribute, which is
        # present on base
        nir09 = dehydrated["assets"]["nir09"]
        assert list(nir09.keys()) == ["href", "eo:bands", "raster:bands"]
        assert len(nir09["raster:bands"]) == 1
        assert list(nir09["raster:bands"][0].keys()) == ["unit"]
        assert nir09["raster:bands"] == [{"unit": DO_NOT_MERGE_MARKER}]

    def test_single_depth_equals(self) -> None:
        base_item = {"a": "first", "b": "second", "c": "third"}
        item = {"a": "first", "b": "second", "c": "third"}
        dehydrated = self.dehydrate(base_item, item)
        assert dehydrated == {}

    def test_nested_equals(self) -> None:
        base_item = {"a": "first", "b": "second", "c": {"d": "third"}}
        item = {"a": "first", "b": "second", "c": {"d": "third"}}
        dehydrated = self.dehydrate(base_item, item)
        assert dehydrated == {}

    def test_nested_extra_keys(self) -> None:
        """
        Test that items having nested dicts with keys not in base item preserve
        the additional keys in the dehydrated item.
        """
        base_item = {"a": "first", "b": "second", "c": {"d": "third"}}
        item = {
            "a": "first",
            "b": "second",
            "c": {"d": "third", "e": "fourth", "f": "fifth"},
        }
        dehydrated = self.dehydrate(base_item, item)
        assert dehydrated == {"c": {"e": "fourth", "f": "fifth"}}

    def test_list_of_dicts_extra_keys(self) -> None:
        """Test that an equal length list of dicts is dehydrated correctly."""
        base_item = {"a": [{"b1": 1, "b2": 2}, {"c1": 1, "c2": 2}]}
        item = {"a": [{"b1": 1, "b2": 2, "b3": 3}, {"c1": 1, "c2": 2, "c3": 3}]}

        dehydrated = self.dehydrate(base_item, item)
        assert "a" in dehydrated
        assert dehydrated["a"] == [{"b3": 3}, {"c3": 3}]

    def test_equal_len_list_of_mixed_types(self) -> None:
        """
        Test that a list of equal length containing matched
        types at each index dehydrates
        dicts and preserves item-values of other types.
        """
        base_item = {"a": [{"b1": 1, "b2": 2}, "foo", {"c1": 1, "c2": 2}, "bar"]}
        item = {
            "a": [
                {"b1": 1, "b2": 2, "b3": 3},
                "far",
                {"c1": 1, "c2": 2, "c3": 3},
                "boo",
            ],
        }

        dehydrated = self.dehydrate(base_item, item)
        assert "a" in dehydrated
        assert dehydrated["a"] == [{"b3": 3}, "far", {"c3": 3}, "boo"]

    def test_unequal_len_list(self) -> None:
        """Test that unequal length lists preserve the item value exactly."""
        base_item = {"a": [{"b1": 1}, {"c1": 1}, {"d1": 1}]}
        item = {"a": [{"b1": 1, "b2": 2}, {"c1": 1, "c2": 2}]}

        dehydrated = self.dehydrate(base_item, item)
        assert "a" in dehydrated
        assert dehydrated["a"] == item["a"]

    def test_marked_non_merged_fields(self) -> None:
        base_item = {"a": "first", "b": "second", "c": {"d": "third", "e": "fourth"}}
        item = {
            "a": "first",
            "b": "second",
            "c": {"d": "third", "f": "fifth"},
        }
        dehydrated = self.dehydrate(base_item, item)
        assert dehydrated == {"c": {"e": DO_NOT_MERGE_MARKER, "f": "fifth"}}

    def test_marked_non_merged_fields_in_list(self) -> None:
        base_item = {
            "a": [{"b": "first", "d": "third"}, {"c": "second", "e": "fourth"}],
        }
        item = {"a": [{"b": "first"}, {"c": "second", "f": "fifth"}]}

        dehydrated = self.dehydrate(base_item, item)
        assert dehydrated == {
            "a": [
                {"d": DO_NOT_MERGE_MARKER},
                {"e": DO_NOT_MERGE_MARKER, "f": "fifth"},
            ],
        }

    def test_deeply_nested_dict(self) -> None:
        base_item = {"a": {"b": {"c": {"d": "first", "d1": "second"}}}}
        item = {"a": {"b": {"c": {"d": "first", "d1": "second", "d2": "third"}}}}

        dehydrated = self.dehydrate(base_item, item)
        assert dehydrated == {"a": {"b": {"c": {"d2": "third"}}}}

    def test_equal_list_of_non_dicts(self) -> None:
        """Values of lists that match base_item should be dehydrated off."""
        base_item = {"assets": {"thumbnail": {"roles": ["thumbnail"]}}}
        item = {
            "assets": {"thumbnail": {"roles": ["thumbnail"], "href": "http://foo.com"}},
        }

        dehydrated = self.dehydrate(base_item, item)
        assert dehydrated == {"assets": {"thumbnail": {"href": "http://foo.com"}}}

    def test_invalid_assets_marked(self) -> None:
        """
        Assets can be included on item-assets that are not uniformly included on
        individual items. Ensure that base item asset keys without a matching item
        key are marked do-no-merge after dehydration.
        """
        base_item = {
            "type": "Feature",
            "assets": {
                "asset1": {"name": "Asset one"},
                "asset2": {"name": "Asset two"},
            },
        }
        hydrated = {
            "assets": {"asset1": {"name": "Asset one", "href": "http://foo.com"}},
        }

        dehydrated = self.dehydrate(base_item, hydrated)

        assert dehydrated == {
            "type": DO_NOT_MERGE_MARKER,
            "assets": {
                "asset1": {"href": "http://foo.com"},
                "asset2": DO_NOT_MERGE_MARKER,
            },
        }

    def test_top_level_base_keys_marked(self) -> None:
        """
        Top level keys on the base item not present on the incoming item should
        be marked as do not merge, no matter the nesting level.
        """
        base_item = {
            "single": "Feature",
            "double": {"nested": "value"},
            "triple": {"nested": {"deep": "value"}},
            "included": "value",
        }
        hydrated = {"included": "value", "unique": "value"}

        dehydrated = self.dehydrate(base_item, hydrated)

        assert dehydrated == {
            "single": DO_NOT_MERGE_MARKER,
            "double": DO_NOT_MERGE_MARKER,
            "triple": DO_NOT_MERGE_MARKER,
            "unique": "value",
        }
